from setuptools import setup, find_packages

setup(
    name='yamlito',
    version='0.1.0',
    packages=find_packages(),
    url='https://github.com/hzanoli/yamlito',
    license='GNU Lesser General Public License v3.0',
    author='Henrique Zanoli',
    author_email='hzanoli@gmail.com',
    description='Simple yaml config loader',
    install_requires=['PyYAML'],
    test_requires=['pytest', 'pytest-cov', 'coverage'],
    test_suite='pytest'
)



import io
import pytest

import yamlito.config as config

simple_custom = 'Charm: \n' \
                '  label: Charmino\n'


@pytest.fixture
def default_yaml():
    return io.StringIO("Charm: \n"
                       "  label: Charm Quark\n"
                       "  mass: 1.275\n"
                       "  brothers: \n"
                       "    favorite: [strange, bottom]\n"
                       "    evil_twin: anti-charm\n")


@pytest.fixture()
def default_parsed_dict():
    return {'Charm':
                {'label': 'Charm Quark',
                 'mass': 1.275,
                 'brothers':
                     {'favorite': ['strange', 'bottom'],
                      'evil_twin': 'anti-charm'}}}


@pytest.fixture
def custom_user_dict():
    return {'Charm':
                {'brothers':
                     {'favorite': ['up', 'down']}}}


@pytest.fixture
def custom_user_yaml():
    return io.StringIO("Charm: \n"
                       "  brothers: \n"
                       "    favorite: ['up', 'down']\n")


@pytest.fixture
def custom_user_parsed_dict():
    return {'Charm':
                {'label': 'Charm Quark',
                 'mass': 1.275,
                 'brothers':
                     {'favorite': ['up', 'down'],
                      'evil_twin': 'anti-charm'}}}


def test__update_dict_first_level(default_parsed_dict):
    config._update_dict(default_parsed_dict,
                        {'Charm': {'label': 'Charmino'}})

    correct = {'Charm':
                   {'label': 'Charmino',
                    'mass': 1.275,
                    'brothers':
                        {'favorite': ['strange', 'bottom'],
                         'evil_twin': 'anti-charm'}}}

    assert default_parsed_dict == correct


def test__update_dict_second_level_item(default_parsed_dict):
    config._update_dict(default_parsed_dict,
                        {'Charm':
                             {'brothers':
                                  {'evil_twin': 'c-bar'}}})

    correct = {'Charm':
                   {'label': 'Charm Quark',
                    'mass': 1.275,
                    'brothers':
                        {'favorite': ['strange', 'bottom'],
                         'evil_twin': 'c-bar'}}}

    assert default_parsed_dict == correct


def test__update_dict_second_level_list(default_parsed_dict, custom_user_dict,
                                        custom_user_parsed_dict):
    config._update_dict(default_parsed_dict,
                        custom_user_dict)

    assert default_parsed_dict == custom_user_parsed_dict


def test__read_config_simple_file(default_yaml, default_parsed_dict):
    assert config._read_config_file(default_yaml) == default_parsed_dict


def test__read_config_file_invalid_yaml():
    invalid_yaml_file = io.StringIO("aa:\naaa")
    with pytest.raises(ValueError):
        config._read_config_file(invalid_yaml_file)


def test_read_config_only_default_file(default_yaml, default_parsed_dict):
    assert config.read_config(default_yaml) == default_parsed_dict


def test_read_config_custom_user_file(default_yaml, custom_user_yaml, custom_user_parsed_dict):
    assert config.read_config(default_yaml, custom_user_yaml) == custom_user_parsed_dict


def test_config_default_constructor(default_parsed_dict):
    config_charm = config.Config(**default_parsed_dict["Charm"])
    assert config_charm.label == 'Charm Quark'
    assert config_charm.mass == 1.275
    assert config_charm.brothers.favorite == ['strange', 'bottom']
    assert config_charm.brothers.evil_twin == 'anti-charm'


def test_config_default_parse(default_parsed_dict):
    config_charm = config.Config.parse(default_parsed_dict["Charm"])
    assert config_charm.label == 'Charm Quark'
    assert config_charm.mass == 1.275
    assert config_charm.brothers.favorite == ['strange', 'bottom']
    assert config_charm.brothers.evil_twin == 'anti-charm'


def test_config_default_representation(default_parsed_dict):
    assert repr(config.Config.parse(
        default_parsed_dict)) == " <Charm: <label:'Charm Quark'> <mass:1.275> <brothers: " \
                                 "<favorite:['strange', 'bottom']> <evil_twin:'anti-charm'>>>"


def test_config_to_dict():
    complicated_dict_config = {
        'Charm':
            {'label': 'Charm Quark',
             'mass': 1.275,
             'brothers':
                 {'favorite': ['strange', 'bottom'],
                  'evil_twin': 'anti-charm'}},
        'mystic':
            {'type':
                 {'super_nested':
                      {'hidden': 'nothing',
                       'deeper':
                           {'secret':
                                {'funny': 'over',
                                 'not_so_funny': 13}}},
                  'not_so_nested': ''}}}
    assert config.Config.parse(complicated_dict_config).to_dict() == complicated_dict_config



import collections
import collections.abc
from typing import Union, TextIO

import yaml


def _update_dict(base_dict, keys_to_update):
    for key, value in keys_to_update.items():
        if isinstance(value, collections.abc.Mapping):
            base_dict[key] = _update_dict(base_dict.get(key, {}), value)
        else:
            base_dict[key] = value
    return base_dict


def _read_config_file(stream: Union[TextIO, dict]):
    if isinstance(stream, dict):
        return stream

    try:
        content = yaml.safe_load(stream)
    except yaml.YAMLError as e:
        raise ValueError('Problem reading the yaml file. Error: \n' + str(e)) from None

    return content


def read_config(default: Union[TextIO, dict], custom: Union[TextIO, dict, None] = None) -> dict:
    """
    Reads the configuration stored in a default yaml file, overriding the fields with the ones defined in another
    custom file. The objects passed can be any object accepted by yaml.safe_load or a dict.

    Args:
        default: stream or dict with the the default configuration.
        custom: stream or dict used to override configurations from default

    Returns:
        config_dict: dict with the combined information from default_file and custom_file
    """

    config_dict = _read_config_file(default)

    if custom is not None:
        custom_config = _read_config_file(custom)
        _update_dict(config_dict, custom_config)

    return config_dict


class Config:
    """ Store configurations.
    The configurations are stored as data members, and can be read/written to yaml files.
    The method parse is particularly handy if you store a file with the default configuration and changes it to a custom
    value depending on another file.
    """
    def __init__(self, **options):
        """
        Build the configuration from the kwargs. Nested configurations are supported, and in case a dict of dict is
        passed, the 'child' configuration objects will also be built.
        Args:
            **options: the configuration options. They will be accessible as data members.
        """
        for key, value in options.items():
            if isinstance(value, dict):
                self.__dict__[key] = Config(**value)
            else:
                self.__dict__[key] = value

    @classmethod
    def parse(cls, default: Union[dict, TextIO], custom: Union[dict, TextIO, None] = None):
        """"Build the configuration from a default, overriding with the options from custom.
        Args:
            default: dict or yaml (file-like) with the default configuration
            custom: dict or yaml (file-like) with the custom configuration
        """
        return cls(**read_config(default, custom))

    def __repr__(self):
        return ''.join([' <' + str(k) + ':' + repr(v) + '>' for (k, v) in self.__dict__.items()])

    def to_dict(self) -> dict:
        """
        Export to a dict. Nested configs are exported as dicts as well.
        Returns:
            dict_config: dict with the configuration
        """
        dict_config = dict()
        for key, value in self.__dict__.items():
            if not isinstance(value, Config):
                dict_config[key] = value
            else:
                dict_config[key] = value.to_dict()

        return dict_config

